# -*- coding: utf-8 -*-

# (c) Jérôme Laheurte 2015-2026
# See LICENSE.txt


class PackageInfo: # pylint: disable=too-few-public-methods
    """
    Information about the package
    """

    version = '2.0.0'
    version_info = map(int, version.split('.'))

    project_name = 'ptk'
    project_url = 'https://codeberg.org/fraca7/ptk'
    download_url = f'https://pypi.python.org/packages/source/p/ptk/ptk-{version}.tar.gz'

    author_name = 'J\u00E9r\u00F4me Laheurte'
    author_email = 'jerome@jeromelaheurte.net'

    short_description = 'LR(1) parsing framework for Python'


version = PackageInfo.version  # pylint: disable=invalid-name
version_info = PackageInfo.version_info  # pylint: disable=invalid-name
